from typing import List, Optional, Type

from django.db.models import Model
from django.forms import Form
from django.urls import path, URLPattern, reverse

from ..forms import ToggleableManyToManyFieldForm

from .base import Block
from .structure import CardWidget, HorizontalLayoutBlock


class LabelBlock(Block):
    """
    A ``<label>``

    Args:
        text: the text for the label

    Keyword Args:
        bold: if ``True``, make the label text be bold
        color: the bootstrap/tabler color to assign to the label
        for_input: the CSS id of the input this describes
    """
    tag: str = "label"

    #: If ``True``, make the label text be bold
    bold: bool = True
    #: The CSS id of the input this describes
    for_input: Optional[str] = None

    def __init__(
        self,
        text: str,
        for_input: str = None,
        bold: bool = None,
        color: str = "secondary",
        **kwargs
    ):
        # FIXME: color is a kwarg, but it is not used
        self.bold = bold if bold is not None else self.bold
        self.for_input = for_input if for_input is not None else self.for_input
        super().__init__(text, **kwargs)
        if self.for_input:
            self._attributes['for'] = self.for_input
        if self.bold:
            self.add_class('fw-bold')


class InputBlock(Block):
    """
    A block for rendering a simple ``<input>`` element.

    Example:

        >>> block = InputBlock(input_type='checkbox', input_name='my-checkbox', value=1)

    Keyword Args:
        input_type: the value of the ``type`` attribute
        input_name: the value of the ``name`` attribute
        value: the value of the ``value`` attribute
    """
    empty: bool = True
    tag: str = 'input'

    input_type: Optional[str] = None
    input_name: Optional[str] = None
    value: Optional[str] = None

    def __init__(
        self,
        input_type: str = None,
        input_name: str = None,
        value: str = None,
        **kwargs
    ):
        self.input_type = input_type if input_type else self.input_type
        self.input_name = input_name if input_name else self.input_name
        self.value = value if value else self.value
        super().__init__(**kwargs)
        if self.input_type is not None:
            self._attributes['type'] = self.input_type
        if self.input_name is not None:
            self._attributes['name'] = self.input_name
        if self.value is not None:
            self._attributes['value'] = self.value


class BaseCheckboxInputBlock(InputBlock):
    """
    A block for rendering a bare ``<input type="checkbox">`` element.

    Example:

        >>> block = BaseCheckboxInputBlock(name='my-checkbox', value=1)
        >>> block = BaseCheckboxInputBlock(name='my-checkbox', value=1, checked=true)

    Keyword Args:
        checked: if ``True``, render the checkbox as checked
    """

    input_type: str = 'checkbox'

    def __init__(
        self,
        checked: bool = False,
        **kwargs
    ):
        self.checked = checked
        super().__init__(**kwargs)
        if not self.input_name:
            raise self.RequiredAttrOrKwarg('input_name')
        if not self.value:
            raise self.RequiredAttrOrKwarg('value')
        if self.checked:
            self._attributes['checked'] = ''


class CheckboxInputBlock(Block):
    """
    A block for rendering a ``<input type="checkbox">`` element with a label as a
    `Boostrap 5 check <https://getbootstrap.com/docs/5.2/forms/checks-radios/#checks>`_.

    Example:

        >>> block = CheckboxInputBlock(label_text='My Checkbox', name='my-checkbox', value=1)

    Keyword Args:
        label_text: the text to use for the label
        bold: if ``True``, make the label text be bold
        input_name: the value of the ``name`` attribute
        value: the value of the ``value`` attribute
        checked: if ``True``, render the checkbox as checked
    """

    #: the value of the ``name`` attribute
    input_name: Optional[str] = None
    #: the value of the ``value`` attribute
    value: Optional[str] = None
    #: the text to use for the label
    label_text: Optional[str] = None
    #: if ``True``, make the label text be bold
    bold: bool = True

    def __init__(
        self,
        label_text: str = None,
        bold: bool = None,
        input_name: str = None,
        value: str = None,
        checked: bool = False,
        **kwargs
    ):
        self.label_text = label_text if label_text else self.label_text
        self.input_name = input_name if input_name else self.input_name
        self.bold = bold if bold is not None else self.bold
        self.value = value if value else self.value
        if not self.label_text:
            raise self.RequiredAttrOrKwarg('label_text')
        if not self.input_name:
            raise self.RequiredAttrOrKwarg('input_name')
        if not self.value:
            raise self.RequiredAttrOrKwarg('value')
        self.input_css_id = kwargs.pop('css_id', f'checkbox-{self.input_name}-{self.value}')
        self.checked = checked
        super().__init__(**kwargs)
        self.add_class('form-check')
        self.add_block(
            BaseCheckboxInputBlock(
                input_name=self.input_name,
                value=self.value,
                css_id=self.input_css_id,
                checked=self.checked,
                css_class='form-check-input'
            )
        )
        self.add_block(
            LabelBlock(
                text=self.label_text,
                bold=self.bold,
                for_input=self.input_css_id,
                css_class='form-check-label'
            )
        )


class ToggleSwitchInputBlock(CheckboxInputBlock):
    """
    A block for rendering a ``<input type="checkbox">`` element with a label as
    a `Bootstrap 5 switch <https://getbootstrap.com/docs/5.0/forms/checks-radios/#switches>`_.

    Example:

        >>> block = ToggleSwitchInputBlock(label_text='My Checkbox', name='my-checkbox', value=1)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_class('form-switch')


class HiddenInputBlock(InputBlock):
    """
    A block for rendering a ``<input type="hidden">``.

    Example:

        >>> block = HiddenInputBlock(name='my-checkbox', value=1)
    """

    input_type: str = 'hidden'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.input_name:
            raise self.RequiredAttrOrKwarg('input_name')
        if not self.value:
            raise self.RequiredAttrOrKwarg('value')


class CrispyFormWidget(Block):
    """
    A widget that displays a Crispy Form widget.  This is specifically a Crispy
    Form widget because it uses the ``{% crispy %}`` template tag to render the
    form.

    Note:

        If the form is present in the template context already, (because your
        view inserted it for you) you don't need to supply the :py:attr:`form`
        keyword argument -- :py:class:`CrispyFormWidget` will pick it up
        automatically.

    Keyword Args:
        form: A crispy form to display. If none is specified, it will be assumed
            that ``form`` is specified elsewhere in the context.
    """
    template_name: str = 'wildewidgets/crispy_form_widget.html'
    block: str = "wildewidgets-crispy-form-widget"

    def __init__(self, *blocks, form: Form = None, **kwargs):
        super().__init__(*blocks, **kwargs)
        self.form = form

    def get_context_data(self, *args, **kwargs):
        kwargs = super().get_context_data(*args, **kwargs)
        if self.form:
            kwargs['form'] = self.form
        return kwargs


class ToggleableManyToManyFieldBlock(CardWidget):
    """
    This is a block that allows you to manage a many to many relationship
    between a model and a related model as a list of toggleable items.  The
    block lists all available items, showing which ones are selected on our
    model instance as checked toggle switches.  It also offers a search box
    to search for items in the list, and a control to hide unselected items.

    This is mostly appropriate for relationships between a model and a
    list-of-values type lookup table (e.g. categories, tags, classifiers, etc.).

    Keyword Args:
    """

    name: str = 'toggle-form-block'

    SCRIPT: str = """
document.querySelectorAll("{target} input.form-check-input").forEach(input => {{
    if (!input.checked) {{
        input.parentElement.classList.add('d-none');
    }};
}});
var show_input = document.getElementById("{show_unselected_id}");
show_input.onchange = function(e) {{
    document.querySelectorAll("{target} input.form-check-input").forEach(input => {{
        if (show_input.checked && !input.checked) {{
            input.parentElement.classList.add('d-none');
        }} else {{
            input.parentElement.classList.remove('d-none');
        }};
    }});
}};
var filter_input = document.getElementById("{filter_id}");
filter_input.onkeyup = function(e) {{
    var filter = filter_input.value.toLowerCase();
    show_input.checked = false
    document.querySelectorAll("{target} label").forEach(label => {{
        var test_string = label.innerText.toLowerCase();
        if (test_string.includes(filter)) {{
            label.parentElement.classList.remove('d-none');
        }} else {{
            label.parentElement.classList.add('d-none');
        }}
    }});
}};
"""

    #: The model this widget will be used with.  This is only used by our
    #: :py:meth:`as_v`
    model: Optional[Type[Model]] = None
    #: The name of the related field on our model instance for which we want to
    #: build this widget
    field_name: Optional[str] = None
    #: The form class to instantiate to handle our multiple select
    form_class: Type[Form] = ToggleableManyToManyFieldForm
    #: The URL to which to POST this form
    form_action: Optional[str] = None
    #: The path prefix to use to root view.
    url_prefix: str = ''
    #: The URL namespace to use for our view name.  This should be set to
    #: the ``app_name`` set in the ``urls.py`` where this viewset's patterns
    #: will be added to ``urlpatterns``.
    url_namespace: Optional[str] = None

    def __init__(
        self,
        instance: Model,
        field_name: str = None,
        form_class: Type[Form] = None,
        form_action: str = None,
        **kwargs
    ):
        self.instance = instance
        self.model = instance.__class__
        if hasattr(self.model, 'model_verbose_name_plural'):
            self.verbose_name_plural = self.model.model_verbose_name_plural()
        else:
            self.verbose_name_plural = self.model._meta.verbose_name_plural
            if not self.verbose_name_plural[0].isupper():
                self.verbose_name_plural = self.verbose_name_plural.capitalize()
        self.field_name = field_name if field_name else self.field_name
        self.field = self.instance._meta.get_field(self.field_name)
        self.related_model = self.field.related_model
        self.form_class = form_class if form_class else self.form_class
        self.form_action = form_action if form_action is not None else self.form_action
        if not self.form_action:
            url_name = self.get_url_name()
            if self.url_namespace:
                url_name = f'{self.url_namespace}:{url_name}'
            self.form_action = reverse(url_name, kwargs={'pk': self.instance.id})
        kwargs['script'] = self.SCRIPT.format(
            target=f'#{self.form_id}',
            filter_id=self.filter_id,
            show_unselected_id=self.show_all_switch_id
        )
        super().__init__(self, **kwargs)
        self.set_header(self.get_header)
        self.set_widget(
            CrispyFormWidget(
                form=self.get_form(self.instance, self.field_name, self.form_action),
                css_id=self.form_id
            )
        )

    @property
    def form_id(self) -> str:
        """
        Return the CSS id we should use for our form.
        """
        return f'{self.model._meta.object_name.lower()}_{self.field_name}'

    @property
    def filter_id(self) -> str:
        """
        Return the CSS id we should use for the filter items search input.
        """
        return f'{self.form_id}_filter'

    @property
    def show_all_switch_id(self) -> str:
        """
        Return the CSS id we should for our "Hide unselected" toggle.
        """
        return f'{self.form_id}_show_all'

    def get_form(
        self,
        instance: Model,
        field_name: str,
        form_action: str
    ) -> Form:
        """
        Return the form to render into our widget, bound to ``field_name`` on
        ``instance``.

        Args:
            instance: the model instance to which to bind our form
            field_name: the name of our many-to-many field on our model
            form_action: the URL to which to POST our form

        Returns:
            An instance of :py:attr:`form_class` bound to ``instance``.
        """
        return self.form_class(instance, field_name=field_name, form_action=form_action)

    def get_header(self) -> Block:
        """
        Get our card header.  This consists of a toggle switch which hides/shows
        unselected items and a search input that allows the user to search for
        items.

        Returns:
            The card header block.
        """
        # We have to do this import here due to circular dependencies
        from ..models import model_verbose_name_plural
        return HorizontalLayoutBlock(
            ToggleSwitchInputBlock(
                label_text="Hide unselected",
                input_name='hide-unchecked',
                value='hide',
                css_id=self.show_all_switch_id,
                css_class='pt-3',
                checked=True
            ),
            Block(
                LabelBlock(
                    f'Filter {model_verbose_name_plural(self.related_model)}',
                    css_class="d-none",
                    attributes={
                        'for': self.filter_id,
                    },
                ),
                InputBlock(
                    attributes={
                        'type': 'text',
                        'placeholder': f'Filter {model_verbose_name_plural(self.related_model)}',
                    },
                    css_id=self.filter_id,
                    css_class='form-control',
                ),
                css_class='w-25',
            ),
            justify='between',
            align='center',
            css_class='w-100'
        )

    @classmethod
    def get_url_name(cls) -> str:
        from ..models import model_logger_name
        model_name = model_logger_name(cls.model)
        related_model = cls.model._meta.get_field(cls.field_name).related_model
        related_model_name = model_logger_name(related_model)
        return f'{model_name}--{related_model_name}--update'

    @classmethod
    def get_urlpatterns(cls, url_prefix: str = None, url_namespace: str = None) -> List[URLPattern]:
        """
        Build a view that will service this block and return a
        :py:class:`django.urls.URLPattern` for that view that you can add to
        your ``urlpatterns``.

        Example::

            from typing import List
            from django.urls import path, URLPattern

            from wildewidgets import ToggleableManyToManyFieldBlock

            from .views import HomeView
            from .models import MyModel

            class TagsSelectorWidget(ToggleableManyToManyFieldBlock):

                model = MyModel
                field_name = 'tags'

            app_name: str = "myapp"

            urlpatterns: List[URLPattern] = [
                path('', HomeView.as_view(), name='home'),
            ]
            urlpatterns += TagsSelectorWidget.get_urlpatterns(url_namespace=app_name)

        Important:
            In order for this to work, you must have subclassed
            :py:class:`ToggleableManyToManyFieldBlock` and defined both the
            :py:attr:`model` and :py:attr:`field_name` attributes.

        Keyword Args:
            url_prefix: a prefix to the path we will build for our view
            url_namespace: the namespace for our url pattern.  We'll set our
                :py:attr:`url_namespace` from this

        Returns:
            A list of urlpatterns for a view suitable for this block
        """
        from ..views.generic import ManyToManyRelatedFieldView
        from ..models import model_logger_name
        if cls.model is None:
            raise ValueError('Define the "model" class attribute before calling "get_urlpattern"')
        if cls.field_name is None:
            raise ValueError('Define the "field_name" class attribute before calling "get_urlpattern"')
        model_name = model_logger_name(cls.model)
        related_model = cls.model._meta.get_field(cls.field_name).related_model
        related_model_name = model_logger_name(related_model)
        if url_namespace:
            cls.url_namespace = url_namespace
        if not url_prefix:
            url_prefix = cls.url_prefix
        elif not url_prefix.endswith('/'):
            url_prefix = f'{url_prefix}/'
        view_path = path(
            f'{url_prefix}wildewidgets/{model_name}/<int:pk>/{related_model_name}/',
            ManyToManyRelatedFieldView.as_view(model=cls.model, field_name=cls.field_name),
            name=cls.get_url_name()
        )
        return [view_path]



from typing import Any, Dict, List, Optional, Type

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Submit, Layout, Field, Fieldset, ButtonHolder
from django.forms import (
    CheckboxInput,
    CheckboxSelectMultiple,
    Form,
    HiddenInput,
    ModelForm,
    MultipleChoiceField,
    Textarea,
    modelform_factory
)
from django.db.models import Model
from django.db.models.fields import (
    AutoFieldMixin,
    BooleanField,
    DateField,
    DateTimeField,
    TextField,
)
from django.db.models.fields.related import (
    RelatedField,
    ManyToManyField,
    ManyToManyRel,
    ManyToOneRel
)
try:
    from django_extensions.db.fields import (
        CreationDateTimeField,
        ModificationDateTimeField,
        AutoSlugField
    )
    has_django_extensions = True
except ImportError:
    has_django_extensions = False


class AbstractModelFormBuilder:
    """
    A base class for a :py:class:`django.forms.ModelForm` builder, which
    automatically builds a decent ``ModelForm`` based on a
    :py:class:`django.db.models.Model`.
    """

    @classmethod
    def factory(
        cls,
        model_class: Type[Model],
        form_action: str,
        fields: List[str] = None,
        excludes: List[str] = None,
    ) -> "Type[ModelForm]":
        """
        Construct a form class for the :py:class:`Model` subclass ``model_class``
        and return it.


        Args:
            model_class: the model class for which to build a form
            form_class_name: the name to use for the form class
            form_action: the URL to which to post our form

        Keyword Args:
            excludes: exclude fields named in this list in addition

        Returns:
            A configured form class.
        """


class AbstractRelatedModelFormBuilder:
    """
    A base class for a formbuilder for building a form which allows users to
    manage multi-select ManyToManycontains only a
    :py:class:`django.forms.CheckboxSelectMultiple` widget that allows the user
    to manage a :py:class:`django.db.models.ManyToManyField` from a model to a
    lookup table and vice-versa.
    """

    @classmethod
    def factory(
        cls,
        model_class: Type[Model],
        field_name: str,
        form_action: str
    ) -> "Type[ModelForm]":
        """
        Construct a form class for the :py:class:`Model` subclass ``model_class``
        and return it.


        Args:
            model_class: the model class for which to build a form
            form_class_name: the name to use for the form class
            form_action: the URL to which to post our form

        Keyword Args:
            excludes: exclude fields named in this list in addition

        Returns:
            A configured form class.
        """


class AutoCrispyModelForm(AbstractModelFormBuilder, ModelForm):
    """
    This is a :py:class:`ModelForm` styled with ``django-crispy-forms` widgets
    that automatically configures itself to look like we want based on how our
    ``Meta`` class settings.

    * Set this to be a ``POST`` form
    * Set our form action to :py:attr:`form_action`
    * Style our inputs with Boostrap styling
    * Add a submit button
    """

    #: The URL to which we will post this form
    form_action: str
    #: How much margin we should put vertically between inputs
    vertical_spacing: str = 'mb-2'

    @classmethod
    def factory(
        cls,
        model_class: Type[Model],
        form_action: str,
        fields: List[str] = None,
        excludes: List[str] = None,
    ) -> "Type[AutoCrispyModelForm]":
        """
        Construct a form class for the :py:class:`Model` subclass ``model_class``
        and return it.

        Exclude these kinds of fields from the form:

        * :py:class:`django.db.fields.AutoField`
        * :py:class:`django.db.fields.BigAutoField`
        * :py:class:`django.db.fields.SmallAutoField`
        * :py:class:`django.db.fields.DateField` with ``auto_now == True``
        * :py:class:`django.db.fields.DateTimeField` with ``auto_now == True``

        If you use ``django_extensions``, we also exclude these fields from the form:

        * :py:class:`django_extensions.db.fields.CreationDateTimeField`
        * :py:class:`django_extensions.db.fields.ModificationDateTimeField`
        * :py:class:`django_extensions.db.fields.AutoSlugField`

        Assign widgets for some common use cases:

        * Use :py:class:`django.forms.Textarea` for :py:class:`django.db.fields.TextField` fields
        * Give all :py:class:`dhango.db.fields.BooleanField` fields Bootstrap 5 Switch styling
          (See `Bootstrap 5 Checks and Radios <https://getbootstrap.com/docs/5.2/forms/checks-radios/#switches>`_)

        Note:
            Specify either ``fields`` or ``excludes``, but not both.

        Args:
            model_class: the model class for which to build a form
            form_action: the URL to which to post our form

        Keyword Args:
            fields: include these fields in the form.
            excludes: exclude fields named in this list in addition to ones excluded by type

        Returns:
            A configured form class.
        """
        if fields and excludes:
            raise ValueError('Specify either "fields" or "excludes" but not both')
        excludes = excludes if excludes else []
        excludes = set(excludes)
        widgets = {}
        excluded_types = [
            # All the auto fields are derived from this mixin.  Exclude such
            # fields because humans shouldn't muck with them.
            AutoFieldMixin,
            # Skip related fields -- no easy way to pick them
            RelatedField,
        ]
        if has_django_extensions:
            excluded_types.extend([
                CreationDateTimeField,
                ModificationDateTimeField,
                AutoSlugField,
            ])
        for field in model_class._meta.get_fields():
            if not fields and isinstance(field, tuple(excluded_types)):
                # Exclude any field that should not be user editable
                excludes.add(field.name)
            if not fields and isinstance(field, (DateTimeField, DateField)) and field.auto_now:
                # Exclude any DateField or DateTimeField that sets its time automatically
                excludes.add(field.name)
            elif isinstance(field, TextField):
                # Always make TextFields use Textareas
                widgets[field.name] = Textarea(attrs={'cols': 50, 'rows': 3})
            elif isinstance(field, BooleanField):
                # Make our BooleanFields be switches.  This also requires
                # some work in our :py:meth:`build_fieldset`
                widgets[field.name] = CheckboxInput(attrs={'role': 'switch'})

        form = modelform_factory(
            model_class,
            form=cls,
            fields=fields,
            exclude=list(excludes),
            widgets=widgets
        )
        form.form_action = form_action
        return form

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_class = 'form-horizontal'
        self.helper.label_class = 'col-lg-3'
        self.helper.field_class = 'col'
        self.helper.form_method = 'post'
        self.helper.form_action = self.form_action
        self.helper.layout = Layout(
            self.build_fieldset(),
            ButtonHolder(
                Submit('submit', 'Save', css_class='btn btn-primary'),
                css_class='d-flex flex-row justify-content-end w-100 mt-3'
            )
        )

    def build_fieldset(self) -> Fieldset:
        """
        Build a :py:class:`Fieldset` with all our form fields properly
        configured, and return it.
        """
        fields = ['']
        for name, field in self.fields.items():
            if isinstance(field.widget, CheckboxInput):
                # Make our checkboxes look like Bootstrap 5 switches
                fields.append(
                    Field(
                        name,
                        wrapper_class='form-check form-switch',
                        css_class=self.vertical_spacing
                    )
                )
            else:
                fields.append(Field(name, css_class=self.vertical_spacing))
        return Fieldset(*fields)


class AbstractRelatedFieldForm(Form):
    """
    This abstract class is the basis for creating forms that manage a single
    related field on a model.
    """

    #: The name of the  :py:class:`django.db.models.ManyToManyField` we want to manage
    field_name: Optional[str] = None
    form_action: Optional[str] = None

    def __init__(
        self,
        instance: Model,
        *args,
        field_name: str = None,
        form_action: str = None,
        **kwargs
    ):
        #: The instance whose :py:class:`django.db.models.ManyToManyField` we want to manage
        self.instance = instance
        self.field_name = field_name if field_name else self.field_name
        self.form_action = form_action if form_action else self.form_action
        if self.field_name is None:
            raise ValueError(
                'You must provide "field_name" as either a constructor keyword argument '
                'or as a class attribute'
            )
        super().__init__(*args, **kwargs)
        #: The field instance for our related field
        self.field = instance._meta.get_field(self.field_name)
        if not isinstance(self.field, RelatedField):
            raise ValueError(
                f'{instance._meta.object_name}.{self.field_name} is not a related field. '
                f'It is instead a {self.field.__class__.__name__}.'
            )
        #: The related model for the related field
        self.related_model = self.field.related_model
        self.pk_field_name = instance._meta.pk.name
        self.fields = self.get_fields()

        self.helper = FormHelper()
        self.helper.form_class = 'form'
        self.helper.form_method = 'post'
        self.helper.form_show_labels = False
        self.helper.form_action = self.form_action
        self.helper.layout = self.get_fieldset()

    def get_fields(self) -> Dict[str, Any]:
        return {}

    def get_fieldset(self) -> Fieldset:
        return Layout(
            Fieldset(
                '',
                Field(self.field_name, id=self.field_name)
            ),
            ButtonHolder(
                Submit('submit', 'Update', css_class='btn btn-primary'),
                css_class='d-flex flex-row justify-content-end button-holder'
            )
        )

    def save(self) -> Model:
        setattr(self.instance, self.cleaned_data[self.field_name])
        return self.instance


class ToggleableManyToManyFieldForm(AbstractRelatedFieldForm):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not isinstance(self.field, (ManyToManyField, ManyToManyRel)):
            raise ValueError(
                f'{self.instance._meta.object_name}.{self.field_name} is not a ManyToManyField or many-to-many reverse '
                f'relation.  Instead, it is a {self.field.__class__.__name__}'
            )

    def get_fields(self) -> Dict[str, Any]:
        fields = super().get_fields()
        choices = [(r.pk, str(r)) for r in self.related_model.objects.all()]
        related_pk_name = self.related_model._meta.pk.name
        initial = list(
            getattr(self.instance, self.field_name)
            .values_list(related_pk_name, flat=True)
        )
        fields[self.field_name] = MultipleChoiceField(
            choices=choices,
            initial=initial,
            required=False,
            widget=CheckboxSelectMultiple(attrs={"class": "form-control"})
        )
        return fields

    def get_fieldset(self) -> Fieldset:
        return Layout(
            Fieldset(
                '',
                Field(
                    self.field_name,
                    id=self.field_name,
                    wrapper_class='form-check form-switch form-check-reverse',
                ),
            ),
            ButtonHolder(
                Submit('submit', 'Update', css_class='btn btn-primary'),
                css_class='d-flex flex-row justify-content-end button-holder'
            )
        )

    def save(self) -> Model:
        getattr(self.instance, self.field_name).set(self.cleaned_data[self.field_name])
        return self.instance



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Dict, Any, Optional, Union

from .base import Block, TemplateWidget, Image
from .structure import HorizontalLayoutBlock
from .text import CodeWidget


class KeyValueListBlock(Block):

    tag: str = 'ul'
    block: str = "list-group"

    def add_simple_key_value(self, key, value):
        self.add_block(
            HorizontalLayoutBlock(
                Block(key),
                Block(value),
                tag="li",
                css_class="list-group-item",
            )
        )

    def add_code_key_value(self, key, value, language=None):
        self.add_block(
            Block(
                Block(key),
                CodeWidget(
                    code=value,
                    language=language,
                    css_class="m-3",
                ),
                tag="li",
                css_class="list-group-item",
            )
        )


class GravatarWidget(Image):

    block: str = 'rounded-circle'

    #: The gravatar URL
    gravatar_url: Optional[str] = None
    #: The length in pixels that will used as the height and width of the image
    size: Union[int, str] = 28
    #: The person's name.  This will be used as the ``alt`` tag on the image
    fullname: Optional[str] = None

    def __init__(
        self,
        gravatar_url: str = None,
        size: Union[int, str] = None,
        fullname: str = None,
        **kwargs
    ):
        self.gravatar_url = gravatar_url if gravatar_url is not None else self.gravatar_url
        self.size = size if size is not None else self.size
        self.fullname = fullname if fullname is not None else self.fullname
        try:
            int(self.size)
        except ValueError as e:
            raise ValueError(f'size should be an integer; got "{self.size}" instead') from e
        kwargs['src'] = self.gravatar_url
        if self.fullname:
            kwargs['alt'] = self.fullname
        super().__init__(**kwargs)
        self._attributes['style'] = f'width: {self.size}px; height: {self.size}px'


class InitialsAvatarWidget(TemplateWidget):

    template_name: str = 'wildewidgets/initials_avatar.html'

    #: The length in pixels that will used as the height and width of the gravatar
    size: int = 28
    #: The foreground color for the gravatar
    color: str = "white"
    #: The background color for the gravatar
    background_color: str = "#626976"

    def __init__(
        self,
        *args,
        initials: str = None,
        size: int = None,
        color: str = None,
        background_color: str = None,
        fullname: str = None,
        **kwargs
    ):
        self.initials = initials
        self.size = size or self.size
        self.color = color or self.color
        self.background_color = background_color or self.background_color
        self.fullname = fullname
        super().__init__(*args, **kwargs)

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        kwargs = super().get_context_data(**kwargs)
        kwargs['initials'] = self.initials.upper()
        kwargs['fullsize'] = str(self.size)
        kwargs['halfsize'] = str(self.size / 2)
        kwargs['color'] = self.color
        kwargs['background_color'] = self.background_color
        kwargs['fullname'] = self.fullname
        return kwargs



from copy import deepcopy
import logging
from typing import Any, Dict, List, Optional, Type

from braces.views import (
    FormInvalidMessageMixin,
    FormValidMessageMixin,
    LoginRequiredMixin,
    MessageMixin
)
from django.forms import BaseModelForm, Form
from django.http.response import HttpResponse
from django.shortcuts import redirect
from django.views.generic import (
    CreateView as DjangoCreateView,
    TemplateView,
    UpdateView as DjangoUpdateView,
    View
)
from django.views.generic.edit import (
    BaseDeleteView,
    BaseUpdateView,
)

from ..forms import AbstractRelatedFieldForm, ToggleableManyToManyFieldForm
from ..models import (
    ViewSetMixin,
    model_logger_name,
    model_name,
    model_verbose_name,
    model_verbose_name_plural,
)
from ..views import WidgetInitKwargsMixin, TableActionFormView
from ..widgets import (
    BaseDataTable,
    Block,
    BreadcrumbBlock,
    CardWidget,
    CrispyFormWidget,
    LookupModelTable,
    NavbarMixin,
    RowDeleteButton,
    RowEditButton,
    TwoColumnLayout,
    Widget,
    WidgetListLayoutHeader
)

from .mixins import StandardWidgetMixin
from .permission import PermissionRequiredMixin


logger = logging.getLogger(__name__)


# --------------------------------------------
# Abastract Widgets
# --------------------------------------------

class AbstractFormPageLayout(Block):
    """
    This is a abstact class for making whole page layouts for a form page:
    creating or updating a model.

    It defines a page title, a page subtitle and a block to contain our
    main page content (typically just our model form).

    Subclass this and override :py:meth:`get_form` to render your main
    page content.

    Note:
        This class will try to supply a reasonable :py:attr:`title` and
        :py:attr:`subtitle`, by inferring reasonable values based on what base
        classes are in the ``view`` class hierarchy.  So you only need to set
        those if you don't get a title/subtitle that you like.

    Args:
        view: the view instance that will render this layout

    Keyword Args:
        title: use this as our page title
        subitle: Use this as our page subtitle.  This will appear directly
            below the page title.  Typically set this to a string naming the
            object we're updating
    """

    block: str = 'form-layout'

    #: Use this as our page ittle
    title: Optional[str] = None
    #: Use this as our page subtitle.  This will appear directly below
    #: the page title.  Typically set this to a string naming the object
    #: we're updating
    subtitle: Optional[str] = None

    def __init__(
        self,
        view: "GenericViewMixin",
        title: str = None,
        subtitle: str = None,
        **kwargs
    ):
        self.view = view
        self.title = title if title else self.title
        self.subtitle = subtitle if subtitle else self.subtitle
        if not self.title:
            if isinstance(self.view, DjangoCreateView):
                self.title = f'Create {self.view.model_verbose_name}'
            elif isinstance(self.view, DjangoUpdateView):
                self.title = f'Update {self.view.model_verbose_name}'
            else:
                self.title = f'Manage {self.view.model_verbose_name}'
        if not self.subtitle and view.object is not None:
            self.subtitle = str(view.object)
        super().__init__(**kwargs)
        self.add_block(self.get_title())
        if self.subtitle:
            self.add_block(self.get_subtitle())
        self.add_block(self.get_form())

    def get_title(self) -> Block:
        title = Block(
            self.title,
            name=f'{self.block}__title',
            tag='h1',
        )
        if not self.subtitle:
            title.add_class('mb-5')
        return title

    def get_subtitle(self) -> Optional[Block]:
        return Block(
            self.subtitle,
            name=f'{self.block}__subtitle',
            css_class='text-muted fs-6 text-uppercase mb-5'
        )

    def get_form(self) -> Block:
        raise NotImplementedError


# --------------------------------------------
# Widgets
# --------------------------------------------

class IndexTableWidget(Block):
    """
    This is the block that we use to render the listing table on the
    :py:class:`IndexView`.  It adds a header above the table.
    """

    def __init__(
        self,
        instance: ViewSetMixin,
        table: BaseDataTable,
        show_create_button: bool = True,
        **kwargs
    ):
        self.instance = instance
        self.show_create_button = show_create_button
        super().__init__(**kwargs)
        self.add_block(self.get_title())
        self.add_block(CardWidget(widget=table))

    def get_title(self) -> WidgetListLayoutHeader:
        header = WidgetListLayoutHeader(
            header_text=self.instance._meta.verbose_name_plural.capitalize(),
            badge_text=self.instance.objects.count(),
        )
        if self.show_create_button:
            header.add_link_button(
                text=f"New {self.instance._meta.verbose_name.capitalize()}",
                color="primary",
                url=self.instance.get_create_url()
            )
        return header


class FormPageLayout(AbstractFormPageLayout):

    left_column_width: int = 9

    def get_form(self) -> Block:
        layout = TwoColumnLayout(left_column_width=9)
        form_widget = CrispyFormWidget(
            form=self.view.get_form(),
            css_class='shadow bg-white p-4'
        )
        layout.add_to_left(form_widget)
        return layout


# --------------------------------------------
# View mixins
# --------------------------------------------

class GenericViewMixin:
    """
    This class based view mixin adds some methods and attributes that our
    other generic views use.
    """

    #: Set this to the model permissions the user must have in order
    #: to be authorized for this view:
    required_model_permissions: List[str] = []
    #: Set this to the logger of your choice
    logger: Any = logger
    #: Set this to your :py:class:`wildewidgets.widgets.navigation.BreadcrumbBlock`
    #: to manage breadcrumbs automatically.
    breadcrumbs: Optional[BreadcrumbBlock] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.breadcrumbs is not None:
            self.breadcrumbs = deepcopy(self.breadcrumbs)

    @property
    def model_name(self) -> str:
        """
        Return our model name.
        """
        if hasattr(self.model, 'model_name'):
            return self.model.model_name()
        return model_name(self.model)

    @property
    def model_logger_name(self) -> str:
        """
        Return a string we can use to identify this model in a log message.
        """
        if hasattr(self.model, 'model_logger_name'):
            return self.model.model_logger_name()
        return model_logger_name(self.model)

    @property
    def model_verbose_name(self) -> str:
        """
        Return a string we can use to identify this model in a log message.
        """
        if hasattr(self.model, 'model_verbose_name'):
            return self.model.model_verbose_name()
        return model_verbose_name(self.model)

    @property
    def model_verbose_name_plural(self) -> str:
        if hasattr(self.model, 'model_verbose_name_plural'):
            return self.model.model_verbose_name_plural()
        return model_verbose_name_plural(self.model)

    def get_client_ip(self) -> str:
        """
        Return our user's IP address.  This will either be the leftmost
        IP address in our ``X-Forwarded-For`` header, or the ``REMOTE_ADDR``
        header.

        Returns:
            Our user's IP address
        """
        x_forwarded_for = self.request.META.get('HTTP_X_FORWARDED_FOR')
        if x_forwarded_for:
            return x_forwarded_for.split(',')[0]
        return self.request.META.get('REMOTE_ADDR')

    @property
    def logging_extra(self) -> str:
        """
        Build some extra stuff to append to our logging statements:

        * add the user's remote_ip
        * add the user's username
        * add the pk of the object, if any

        Returns:
            A string to append to our log message
        """
        data = {}
        data['remote_ip'] = self.get_client_ip()
        data['username'] = self.request.user.username
        if hasattr(self, 'object'):
            data['pk'] = self.object.pk
        return ' '.join([f'{key}={value}' for key, value in data.items()])

    def get_menu_item(self) -> Optional[str]:
        """
        Return the name of the main menu item to set active, which is the
        the appropriately capitalized ``verbose_name_plural`` for our model.

        Returns:
            The name of the main menu item to set active.
        """
        return self.model_verbose_name_plural

    def get_permissions_required(self) -> List[str]:
        """
        Overrides :py:meth:`wildewidgets.views.permission.PermissionRequiredMixin.get_permissions_required`.

        Expand the permissions listed in :py:attr:`required_model_permissions`
        into full Django permission strings (e.g. ``appname.add_modelname`` and
        add them to the list of permissions already defined by our superclass.

        Returns:
            The list of Django user permission strings
        """
        required_perms = set(super().get_permissions_required())
        required_perms.update(
            self.get_model_permissions(self.model, self.required_model_permissions)
        )
        return list(required_perms)

    def get_breadcrumbs(self) -> "Optional[BreadcrumbBlock]":
        """
        Add our title to the breadcrumbs and return the updated block.

        Returns:
            The breadcrumb block to use
        """
        if self.breadcrumbs:
            self.breadcrumbs.add_breadcrumb(title=self.get_title())
        return self.breadcrumbs


class GenericDatatableMixin:

    model: Optional[Type[ViewSetMixin]]

    #: The :py:class:`wildewidgets.widgets.tables.base.BaseDataTable` subclass
    #: to use for the listing table
    table_class: Type[BaseDataTable] = LookupModelTable
    #: A dictionary that we will use as the ``**kwargs`` for the constructor of
    #: :py:attr:`table_class`
    table_kwargs: Dict[str, Any] = {
        'striped': True,
        'page_length': 25,
        'buttons': True,
    }

    def get_table_kwargs(self) -> Dict[str, Any]:
        """
        Return a dict that will be used as the ``**kwargs`` for constructing our
        :py:attr:`table_class` instance.

        If there are no bulk_action form actions defined, define one for bulk
        deleting rows.

        Returns:
            Constructor kwargs for :py:class:`table_class`
        """
        table_kwargs = self.table_kwargs
        if 'form_actions' not in table_kwargs and self.bulk_action_url_name:
            table_kwargs['form_actions'] = [
                ('delete', f'Delete {self.model_verbose_name}')
            ]

        return table_kwargs

    def get_table(self, *args, **kwargs) -> BaseDataTable:
        kwargs.update(self.get_table_kwargs())
        kwargs['model'] = self.model
        actions = []
        if self.user_can_update():
            actions.append(RowEditButton())
        if self.user_can_delete():
            actions.append(RowDeleteButton())
        if actions:
            kwargs['actions'] = actions
        return self.table_class(*args, **kwargs)


# --------------------------------------------
# Views
# --------------------------------------------

class IndexView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    GenericViewMixin,
    GenericDatatableMixin,
    NavbarMixin,
    StandardWidgetMixin,
    TemplateView
):

    #: The model we're describing in this view. It must have
    #: :py:class:`wildewidgets.models.ViewSetMixin` in its class heirarchy
    model: Optional[Type[ViewSetMixin]] = None

    required_model_permissions: List[str] = ['view', 'add', 'change', 'delete']

    def get_title(self) -> str:
        """
        Return our page title.

        Returns:
            The string to use as our page title.
        """
        return self.model_verbose_name_plural

    def get_content(self) -> Widget:
        return IndexTableWidget(
            self.model,
            self.get_table(),
            show_create_button=self.user_can_create()
        )


class TableAJAXView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    WidgetInitKwargsMixin,
    GenericDatatableMixin,
    View
):
    #: The model we're describing in this view. It must have
    #: :py:class:`wildewidgets.models.ViewSetMixin` in its class heirarchy
    model: Optional[Type[ViewSetMixin]] = None

    required_model_permissions: List[str] = ['view', 'add', 'change', 'delete']

    def dispatch(self, request, *args, **kwargs):
        csrf_token = request.GET.get('csrf_token', '')
        extra_data = self.get_decoded_extra_data(request)
        initargs = extra_data.get('args', [])
        initkwargs = extra_data.get('kwargs', {})
        table = self.get_table(*initargs, **initkwargs)
        table.request = request
        table.csrf_token = csrf_token
        table.args = initargs
        table.kwargs = initkwargs
        return table.dispatch(request, *args, **kwargs)


class TableBulkActionView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    GenericViewMixin,
    GenericDatatableMixin,
    MessageMixin,
    TableActionFormView
):
    #: The model we're describing in this view. It must have
    #: :py:class:`wildewidgets.models.ViewSetMixin` in its class heirarchy
    model: Optional[Type[ViewSetMixin]] = None

    #: Set this to the logger of your choice
    logger: Any = logger

    required_model_permissions: List[str] = ['delete']

    def process_delete_action(self, items: List[str]) -> None:
        qs = self.model.objects.filter(id__in=items)
        count = qs.count()
        qs.delete()
        self.logger.info(
            '%s.bulk.delete ids=%s %s',
            self.model_logger_name,
            ','.join(items),
            self.logging_extra
        )
        self.messages.success(f'Deleted {count} {self.model.model_verbose_name_plural()}.')


class CreateView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    FormInvalidMessageMixin,
    FormValidMessageMixin,
    GenericViewMixin,
    NavbarMixin,
    StandardWidgetMixin,
    DjangoCreateView
):

    required_model_permissions: List[str] = ['add']

    #: Use this :py:class:`AbstractFormPageLayout` subclass to render our page
    layout_widget: Type[AbstractFormPageLayout] = FormPageLayout

    def get_form_invalid_message(self):
        self.logger.warning(
            '%s.create.failed.validation %s',
            self.model_logger_name,
            self.logging.extra
        )
        return f"Couldn't create this {self.model_verbose_name} due to validation errors; see below."

    def get_form_valid_message(self):
        self.logger.info('%s.create.success %s', self.model_logger_name, self.logging_extra)
        return f'Created {self.model_verbose_name} "{str(self.object)}"!'

    def get_form_class(self) -> Type[BaseModelForm]:
        # FIXME: this is probably not the right way to do this.  I'd like it
        # that if there is a form_class defined on us, use that otherwise ask
        # the model
        if hasattr(self.model, 'get_create_form_class'):
            return self.model.get_create_form_class()
        return super().get_form_class()

    def get_title(self) -> str:
        """
        Return our page title.

        Returns:
            The string to use as our page title.
        """
        return f'Create {self.model_verbose_name}'

    def get_content(self) -> Widget:
        return self.layout_widget(self, modifier='create')


class UpdateView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    FormInvalidMessageMixin,
    FormValidMessageMixin,
    GenericViewMixin,
    NavbarMixin,
    StandardWidgetMixin,
    DjangoUpdateView
):

    required_model_permissions: List[str] = ['change']

    #: Use this :py:class:`AbstractFormPageLayout` subclass to render our page
    layout_widget: Type[AbstractFormPageLayout] = FormPageLayout

    def get_form_invalid_message(self):
        self.logger.warning(
            '%s.update.failed.validation %s',
            self.model_logger_name,
            self.logging.extra
        )
        return f"Couldn't update this {self.model_verbose_name} due to validation errors; see below."

    def get_form_valid_message(self):
        self.logger.info('%s.update.success %s', self.model_logger_name, self.logging_extra)
        return f'Updated {self.model_verbose_name} "{str(self.object)}"!'

    def get_form_class(self) -> Type[BaseModelForm]:
        # FIXME: this is probably not the right way to do this.  I'd like it
        # that if there is a form_class defined on us, use that otherwise ask
        # the model
        if hasattr(self.object, 'get_update_form_class'):
            return self.object.get_update_form_class()
        return super().get_form_class()

    def get_title(self) -> str:
        """
        Return our page title.

        Returns:
            The string to use as our page title.
        """
        return f'Update {self.model_verbose_name} "{str(self.object)}"'

    def get_content(self) -> Widget:
        return self.layout_widget(self, modifier='update')


class DeleteView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    GenericViewMixin,
    FormInvalidMessageMixin,
    FormValidMessageMixin,
    BaseDeleteView
):

    model: Optional[Type[ViewSetMixin]] = None

    http_method_names: List[str] = ['post']

    required_model_permissions: List[str] = ['delete']

    def get_form_invalid_message(self):
        self.logger.warning('%s.delete.failed.validation', self.model_logger_name)
        return f'Couldn\'t delete {self.model_verbose_name} "{str(self.object)}"'

    def get_form_valid_message(self):
        self.logger.info('%s.delete.success %s', self.model_logger_name, self.logging_extra)
        return f'Deleted {self.model_verbose_name} "{str(self.object)}"!'

    def form_invalid(self, form: BaseModelForm) -> HttpResponse:
        return redirect(self.get_success_url())

    def get_form_class(self) -> Type[BaseModelForm]:
        # FIXME: this is probably not the right way to do this.  I'd like it
        # that if there is a form_class defined on us, use that otherwise ask
        # the model
        if hasattr(self.object, 'get_delete_form_class'):
            return self.object.get_delete_form_class()
        return super().get_form_class()


class ManyToManyRelatedFieldView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    FormValidMessageMixin,
    MessageMixin,
    GenericViewMixin,
    BaseUpdateView
):

    field_name: Optional[str] = None
    form_class: Type[AbstractRelatedFieldForm] = ToggleableManyToManyFieldForm

    required_model_permissions: List[str] = ['change']

    @property
    def model_logger_name(self) -> str:
        return f'{super().model_logger_name}.{self.field_name}'

    @property
    def field_verbose_name(self) -> str:
        return self.model._meta.get_field(self.field_name).verbose_name.capitalize()

    def get_form_kwargs(self) -> Dict[str, Any]:
        kwargs = super().get_form_kwargs()
        kwargs['field_name'] = self.field_name
        return kwargs

    def form_invalid(self, form: Form) -> HttpResponse:
        self.logger.warning(
            '%s.update.failed.validation %s',
            self.model_logger_name,
            self.logging.extra
        )
        self.messages.error(
            f'Couldn\'t update {self.field_verbose_name} on {self.model_verbose_name} "{str(self.object)}"'
        )
        for k, errors in form.errors.as_data().items():
            for error in errors:
                self.messages.error(f"{k}: {error.message}")
        return redirect(self.get_success_url())

    def get_form_valid_message(self):
        self.logger.info('%s.update.success %s', self.model_logger_name, self.logging_extra)
        return f'Updated {self.field_verbose_name} on {self.model_verbose_name} "{str(self.object)}"!'

    def get_success_url(self) -> str:
        success_url = super().get_success_url()
        if not success_url:
            if hasattr(object, 'get_update_url'):
                return self.object.get_update_url()
            return self.object.get_absolute_url()
        return success_url


class HTMXView(TemplateView):
    """
    Simple view to display only a widget as a response. This is useful for
    HTMX requests. The widget is rendered with the context data from the
    request, such that anything sent via GET or POST is passed as kwargs to
    the widget's constructor.

    Example:

        class MyWidgetHTMXView(HTMXView):
            widget_class = MyWidget
    """
    template_name = 'wildewidgets/htmx_base.html'
    widget_class = None
    include_request = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.data = None

    def get_content(self, **kwargs):
        if self.widget_class:
            if self.include_request:
                self.data['request'] = self.request
            return self.widget_class(**self.data)
        raise NotImplementedError

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['content'] = self.get_content(**self.data)
        return context

    def set_data(self, request_data):
        self.data = request_data.dict()
        del self.data['submit']

    def get(self, request, *args, **kwargs):
        self.set_data(request.GET)
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.set_data(request.POST)
        return self.get(request, *args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .base import *  # noqa: F403,F401
from .buttons import *  # noqa: F403,F401
from .charts import *  # noqa: F403,F401
from .datagrid import *  # noqa: F403,F401
from .forms import *  # noqa: F403,F401
from .grid import *  # noqa: F403,F401
from .headers import *  # noqa: F403,F401
from .icons import *  # noqa: F403,F401
from .layout import *  # noqa: F403,F401
from .misc import *  # noqa: F403,F401
from .modals import *  # noqa: F403,F401
from .navigation import *  # noqa: F403,F401
from .structure import *  # noqa: F403,F401
from .tables import *  # noqa: F403,F401
from .text import *  # noqa: F403,F401



"""
Generic view permission mixins
------------------------------

Here we put the classes that derive from
:py:class:`braces.views._access.AccessMixin` because ``AccessMixin` needs Django
to be fully booted in order to be importable.  This is because the module that  ``AccessAdmin``
is in imports some things from ``django.contrib.auth`` that want to load app configs, which
aren't available until after django has fully booted.

For the same reason, don't import this file in ``wildewidgets.views.__init.py``,
because anyone importing any widget that is also a JSON view will cause
``AppRegistryNotReady`` exceptions.
"""
from typing import Optional, List, Type

from braces.views._access import AccessMixin
from django.contrib.auth.models import Group
from django.core.exceptions import PermissionDenied
from django.db.models import Model
from django.http.request import HttpRequest


class PermissionRequiredMixin(AccessMixin):
    """
    This class-based view mixin grants or denies access to a view based on a
    user's group membership or specific permisisons.

    * Django superusers are always authorized
    * Anonymous users are never authorized
    * If ``request.user`` is in one of the groups returned by
      :py:meth:`get_groups_required`, they are authorized.
    * Failing that, if ``request.user`` has one of the permissions returned by
      :py:meth:`get_permissions_required`, they are authorized.
    * Additionally, if :py:attr:`model` and :py:attr:`required_model_permissions`
      are both set, extend what we got from :py:meth:`get_permissions_required`
      with what we get from :py:meth:`get_model_permissions`
    """

    #: Users who are members of the named Django auth groups will be permitted
    #: access to the view, and those who are not will be denied access.
    groups_required: Optional[List[str]] = None

    #: Users who have any of the listed Django model permissions
    #: will be permitted access to the view.
    permissions_required: Optional[List[str]] = None

    #: Used in conjunction with :py:attr:`required_model_permissions`,
    #: this is the model we want to act on.  This should be set by our
    #: our subclass, which is why it is just a type declaration here.
    #: This is used by :py:meth:`get_model_permissions`
    model: Optional[Type[Model]]
    #: A list of simple strings like ``view``, ``add``, ``change``, ``delete`` that
    #: represent actions that can be done on a model.  This is used by
    #: py:meth:`get_model_permissions`
    required_model_permissions: Optional[List[str]] = None

    def user_can_view(self) -> bool:
        required_perms = self.get_model_permissions(self.model, ['view'])
        return self.check_permissions(
            groups_required=[],
            permissions_required=required_perms
        )

    def user_can_create(self) -> bool:
        required_perms = self.get_model_permissions(self.model, ['add'])
        return self.check_permissions(
            groups_required=[],
            permissions_required=required_perms
        )

    def user_can_update(self) -> bool:
        required_perms = self.get_model_permissions(self.model, ['change'])
        return self.check_permissions(
            groups_required=[],
            permissions_required=required_perms
        )

    def user_can_delete(self) -> bool:
        required_perms = self.get_model_permissions(self.model, ['delete'])
        return self.check_permissions(
            groups_required=[],
            permissions_required=required_perms
        )

    def get_groups_required(self) -> List[Group]:
        if self.groups_required is None:
            return []
        return Group.objects.filter(name__in=self.groups_required)

    def get_permissions_required(self) -> List[str]:
        if self.permissions_required is None:
            return []
        return self.permissions_required

    def get_model_permissions(self, model: Model, perms: List[str]) -> List[str]:
        """
        Given a list of simple strings like ``view``, ``add``, ``change``, etc.
        and a Django :py:class:`django.db.models.Model`, return the list of actual
        permissions names that would apply to that model, e.g. for a model ``Foo`` in
        the django app ``core``::

            >>> PermissionMixin.get_model_permissions(Foo, ['add', 'change'])
            ['core.add_foo', 'core.change_foo']

        Args:
            model: the model for which to build permission strings
            perms: the list of short permissions

        Returns:
            The fully formed permission strings
        """
        required_perms = set()
        for perm in perms:
            full_perm = f'{self.model._meta.app_label}.{perm}_{self.model._meta.object_name.lower()}'
            required_perms.add(full_perm)
        return list(required_perms)

    def check_permissions(
        self,
        groups_required: List[Group] = None,
        permissions_required: List[str] = None
    ) -> bool:
        """
        Check if the current user is authorized for our view.

        * Django superusers are always authorized
        * Anonymous users are never authorized
        * If ``request.user`` is in one of the groups returned by
          :py:meth:`get_groups_required`, they are authorized.
        * Failing that, if ``request.user`` has one of the permissions returned by
          :py:meth:`get_permissions_required`, they are authorized.
        * Additionally, if :py:attr:`model` and :py:attr:`required_model_permissions`
          are both set, extend what we got from :py:meth:`get_permissions_required`
          with what we get from :py:meth:`get_model_permissions`

        Args:
            request: the current request

        Keyword Args:
            groups_required: test versus these groups instead of what
                :py:meth:`get_groups_required` returns
            permissions_required: test versus these permissions instead of what
                :py:meth:`get_permissions_required` returns

        Returns:
            Return ``True`` if the user is authorized, ``False`` otherwise.
        """
        if groups_required is None:
            groups_required = self.get_groups_required()
        if permissions_required is None:
            permissions_required = self.get_permissions_required()
        if self.required_model_permissions:
            model_permissions = self.get_model_permissions(
                self.model,
                self.required_model_permissions
            )
            permissions_required.extend(model_permissions)
        if self.request.user.is_superuser:
            # Superusers users are always authorized
            return True
        user_groups = self.request.user.groups.values_list("name", flat=True)
        if set(groups_required).intersection(set(user_groups)):
            return True
        for perm in permissions_required:
            if self.request.user.has_perm(perm):
                return True
        return False

    def dispatch(self, request: HttpRequest, *args, **kwargs):
        self.request = request
        if not self.check_permissions():
            raise PermissionDenied
        return super().dispatch(request, *args, **kwargs)



from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Union, cast

from django.core.exceptions import ImproperlyConfigured

from .base import Block


DatagridItemDef = Union["DatagridItem", Tuple[str, str], Tuple[str, str, Dict[str, Any]]]


class DatagridItem(Block):
    """
    This widget implements a `Tabler datagrid-item
    <https://preview.tabler.io/docs/datagrid.html>`_ It should be used with
    :py:class:`Datagrid`.

    Note:
        Unlike :py:class:`wildewidgets.widgets.base.Block`, :py:class:`DatagridItem` requires
        either :py:attr:`contents` to be set, or the block contents to be
        provided as positional arguments.

    Keyword Args:
        title: the ``datagrid-title`` of the ``datagrid-item``
        url: URL to use to turn content into a hyperlink

    Raises:
        ValueError: either the ``title`` was empty, or no contents were provided
        ImproperlyConfigured: :py:attr:`url` was set, and there is more than one block
            in :py:attr:`contents`
    """
    block: str = 'datagrid-item'

    #: the ``datagrid-title`` of the ``datagrid-item``
    title: Optional[str] = None
    #: a URL to use to turn our contents into a hyperlink
    url: Optional[str] = None

    def __init__(
        self,
        *blocks,
        title: str = None,
        url: str = None,
        **kwargs
    ) -> None:
        self.title = title if title else self.title
        if not self.title:
            raise ValueError('"title" is required as either a keyword argument or as a class attribute')
        self.url = url if url else self.url
        super().__init__(*blocks, **kwargs)

    def add_blocks(self) -> None:
        """
        Add our content.

        If :py:attr:`url` is set, and there is only one block in :py:attr:`contents`, wrap
        that block in a :py:class:`wildewidgets.Link`.

        Raises:
            ImproperlyConfigured: :py:attr:`url` was set, and there is more than one block
                in :py:attr:`contents`
        """
        if self.url:
            if len(self.contents) == 1:
                wrapper: Block = Block(
                    self.contents[0],
                    tag='a',
                    attributes={'href': self.url},
                    name='datagrid-content'
                )
            else:
                raise ImproperlyConfigured(
                    f'{self.__class__.__name__}: url should only be set if contents are a single block'
                )
        else:
            if len(self.contents) == 1:
                if not isinstance(self.contents[0], Block):
                    # Wrap the text in a block to allow us to assign the datagrid-content
                    # class to it
                    wrapper = Block(self.contents[0], name='datagrid-content')
                else:
                    wrapper = self.contents[0]
            else:
                wrapper = Block(name='datagrid-content')
                for block in self.contents:
                    wrapper.add_block(block)
        self.add_block(Block(self.title, name='datagrid-title'))
        self.add_block(wrapper)


class Datagrid(Block):
    """
    This widget implements a `Tabler Data grid <https://preview.tabler.io/docs/datagrid.html>`_
    It contains :py:class:`DatagridItem` objects.

    Examples:

        Add :py:class:`DatagridItem` objects to this in one of these ways:

        As constructor arguments::

            >>> item1 = DatagridItem(title='foo', content='bar', url="https://example.com")
            >>> item2 = DatagridItem(title='baz', content='barney')
            >>> item3 = ['foo', 'bar']
            >>> grid = Datagrid(item1, item2, item3)

        By using :py:meth:`add_block` with a :py:class:`DatagridItem`:

            >>> grid = Datagrid(item1, item2, item3)
            >>> grid.add_block(DatagridItem(title='foo', content='bar'))

        By using :py:meth:`add_item`::

            >>> grid = Datagrid(item1, item2, item3)
            >>> grid.add_item('foo', 'bar')

    Args:
        *items: a list of ``datagrid-item`` definitions or :py:class:`DatagridItem` objects.
    """
    block: str = 'datagrid'
    #: a list of ``datagrid-items`` to add to our content
    items: List[DatagridItemDef] = []

    def __init__(self, *items: DatagridItemDef, **kwargs):
        self.items = list(items) if items else deepcopy(self.items)
        super().__init__(**kwargs)
        for item in items:
            if isinstance(item, DatagridItem):
                self.add_block(item)
            elif isinstance(item, tuple):
                if len(item) == 2:
                    item = cast(Tuple[str, str], item)
                    self.add_item(item[0], item[1])
                else:
                    item = cast(Tuple[str, str, Dict[str, Any]], item)
                    self.add_item(item[0], item[1], **item[2])

    def add_item(
        self,
        title: str,
        content: Union[Union[str, Block], List[Union[str, Block]]],
        url: str = None,
        **kwargs
    ) -> None:
        """
        Add a :py:class:`DatagridItem` to our block contents, with
        ``datagrid-title`` of ``title`` and datagrid.

        Examples:

            Start with our grid::

                >>> dg = DataGrid()

            Add a simple key/value::

                >>> dg.add_item('Version', '1.2.3')

            Add a block as the value, and wrap it in a :py:class:`wildewdigets.Link`::

                >>> dg.add_item(
                    'Gravatar',
                    Image(src=static('myapp/images/gravatar.png'), alt='MyGravatar'),
                    url='https://www.google.com'
                )

            Add a list of blocks as the value::

                >>> dg.add_item(
                    'Contributors',
                    [
                        ImageLink(
                            src=static('myapp/images/fred-gravatar.png',
                            alt='Fred'
                            url='https://www.fred.com'
                        ),
                        ImageLink(
                            src=static('myapp/images/barney-gravatar.png',
                            alt='Barney'
                            url='https://www.barney.com'
                        )
                    ],
                    css_class='d-flex flex-row'
                )


        Note:
            To add a :py:class:`DatagridItem` directly, use :py:meth:`add_block`.

        Keyword Args:
            title: the ``datagrid-title`` of the ``datagrid-item``
            content: the ``datagrid-content`` of the ``datagrid-item``
            url: URL to use to turn content into a hyperlink
        """
        if isinstance(content, str) or isinstance(content, Block):
            content = [content]
        self.add_block(DatagridItem(*content, title=title, url=url, **kwargs))



from copy import deepcopy
from typing import Any, Dict, Optional, Type

from django.urls import path
from django.utils.text import slugify
from django.urls import reverse_lazy
from django.views import View

from wildewidgets.views.generic import (
    CreateView,
    DeleteView,
    IndexView,
    UpdateView,
    TableAJAXView,
    TableBulkActionView
)
from wildewidgets.widgets import (
    BaseDataTable,
    BreadcrumbBlock,
    LookupModelTable,
    Navbar
)

from .models import ViewSetMixin


class ModelViewSet:
    """
    A viewset to allow listing, creating, editing and deleting model instances.
    """

    #: The model for which we are building this viewset
    model: Optional[Type[ViewSetMixin]] = None

    #: The path prefix to use to organize our views.
    url_prefix: Optional[str] = None
    #: The URL namespace to use for our view names.  This should be set to
    #: the ``app_name`` set in the ``urls.py`` where this viewset's patterns
    #: will be added to ``urlpatterns``
    url_namespace: Optional[str] = None
    #: Set this to your :py:class:`wildewidgets.widgets.navigation.BreadcrumbBlock`
    #: class to manage breadcrumbs automatically.
    breadcrumbs_class: Optional[Type[BreadcrumbBlock]] = None
    #: The navbar to use.  We'll highlight our verbose name in any of the
    #: menus within the navbar
    navbar_class: Optional[Type[Navbar]] = None
    #: The template name to use for all views
    template_name: Optional[str] = None

    #: The view class to use for the index view; must be a subclass of
    #: :py:class:`wildewidgets.views.generic.IndexView`.
    index_view_class: Type[IndexView] = IndexView
    #: The :py:class:`wildewidgets.widgets.tables.base.BaseDataTable` subclass
    #: to use for the listing table
    index_table_class: Type[BaseDataTable] = LookupModelTable
    #: A dictionary that we will use as the ``**kwargs`` for the constructor of
    #: :py:attr:`table_class`
    index_table_kwargs: Dict[str, Any] = {
        'striped': True,
        'page_length': 50,
        'buttons': True
    }

    #: The view class to use as the AJAX JSON endpoint for the dataTable; must
    #: be a subclass of :py:class:`wildewidgets.views.generic.TableAJAXView`.
    table_ajax_view_class: Type[View] = TableAJAXView

    #: The view class to use for the index view; must be a subclass of
    #: :py:class:`wildewidgets.views.generic.CreateView`.
    table_bulk_action_view_class: Type[View] = TableBulkActionView

    #: The view class to use for the index view; must be a subclass of
    #: :py:class:`wildewidgets.views.generic.CreateView`.
    create_view_class: Type[View] = CreateView

    #: The view class to use for the index view; must be a subclass of
    #: :py:class:`wildewidgets.views.generic.UpdateView`.
    update_view_class: Type[View] = UpdateView

    #: The view class to use for the index view; must be a subclass of
    #: :py:class:`wildewidgets.views.generic.DeleteView`.
    delete_view_class: Type[View] = DeleteView

    def __init__(
        self,
        model: Type[ViewSetMixin] = None,
        url_prefix: str = None,
        url_namespace: str = None,
        **kwargs
    ):
        #: The model we're describing in this view. It must have
        #: :py:class:`wildewidgets.models.ViewSetMixin` in its class heirarchy
        self.model: Type[ViewSetMixin] = model if model else self.model
        # Tell the model we are its viewset
        self.model.viewset = self
        self.name = slugify(self.model._meta.verbose_name.lower())
        self.url_prefix = url_prefix if url_prefix else self.name
        self.url_namespace = url_namespace

        for key, value in kwargs.items():
            setattr(self, key, value)

    def get_url_name(self, view_name: str, namespace: bool = True) -> str:
        """
        Returns the (possibly) namespaced URL name for the given view.

        Args:
            view_name:
        """
        url_name = f'{self.name}--{view_name}'
        if namespace and self.url_namespace:
            url_name = f'{self.url_namespace}:{url_name}'
        return url_name

    def get_breadcrumbs(self, view_name: str) -> Optional[BreadcrumbBlock]:
        breadcrumbs: Optional[BreadcrumbBlock] = None
        if self.breadcrumbs_class:
            breadcrumbs = self.breadcrumbs_class()  # pylint: disable=not-callable
            if view_name != 'index':
                breadcrumbs.add_breadcrumb(
                    title=self.model.model_verbose_name_plural(),
                    url=reverse_lazy(self.get_url_name('index'))
                )
        return breadcrumbs

    def get_index_table_kwargs(self):
        table_kwargs = deepcopy(self.index_table_kwargs)
        if 'form_actions' not in table_kwargs:
            table_kwargs['form_actions'] = [
                ('delete', 'Delete selected')
            ]
        table_kwargs['ajax_url_name'] = self.get_url_name('table-ajax')
        table_kwargs['form_url'] = reverse_lazy(self.get_url_name('table-bulk-actions'))
        return table_kwargs

    @property
    def index_view(self):
        return self.index_view_class.as_view(
            model=self.model,
            template_name=self.template_name,
            table_class=self.index_table_class,
            table_kwargs=self.get_index_table_kwargs(),
            breadcrumbs=self.get_breadcrumbs('index'),
            navbar_class=self.navbar_class
        )

    @property
    def table_ajax_view(self):
        return self.table_ajax_view_class.as_view(
            model=self.model,
            table_class=self.index_table_class,
            table_kwargs=self.get_index_table_kwargs(),
        )

    @property
    def table_bulk_action_view(self):
        return self.table_bulk_action_view_class.as_view(
            model=self.model,
            url=reverse_lazy(self.get_url_name('index')),
            table_class=self.index_table_class,
            table_kwargs=self.get_index_table_kwargs(),
        )

    @property
    def create_view(self):
        return self.create_view_class.as_view(
            model=self.model,
            template_name=self.template_name,
            breadcrumbs=self.get_breadcrumbs('create'),
            navbar_class=self.navbar_class,
            success_url=self.get_url_name('index')
        )

    @property
    def update_view(self):
        return self.update_view_class.as_view(
            model=self.model,
            template_name=self.template_name,
            breadcrumbs=self.get_breadcrumbs('update'),
            navbar_class=self.navbar_class,
            success_url=self.get_url_name('index')
        )

    @property
    def delete_view(self):
        return self.delete_view_class.as_view(
            model=self.model,
            success_url=self.get_url_name('index')
        )

    def get_urlpatterns(self):
        root = f'{slugify(self.model.model_logger_name())}/'
        if self.url_prefix:
            root = f'{self.url_prefix}/{root}'
        return [
            path(root, self.index_view, name=self.get_url_name('index', namespace=False)),
            path(f"{root}table/ajax/", self.table_ajax_view, name=self.get_url_name('table-ajax', namespace=False)),
            path(
                f"{root}table/bulk/",
                self.table_bulk_action_view,
                name=self.get_url_name('table-bulk-actions', namespace=False)
            ),
            path(f"{root}create/", self.create_view, name=self.get_url_name('create', namespace=False)),
            path(f"{root}<int:pk>/", self.update_view, name=self.get_url_name('update', namespace=False)),
            path(
                f"{root}<int:pk>/delete/",
                self.delete_view,
                name=self.get_url_name('delete', namespace=False)
            )
        ]
